from sklearn.linear_model import LinearRegression

from xh_cpu_usage_simulator.MovingAvg import MovingAvg
import numpy as np


class LinearRegressionPredication:
    def __init__(self, size: int):
        self.x_mv = MovingAvg(size)
        self.y_mv = MovingAvg(size)
        self.model = LinearRegression()
        self.predication = None

    def insert(self, x: float, y: float):
        self.x_mv.insert(x)
        self.y_mv.insert(y)

        if not self.is_ready():
            return

        y = np.array(self.y_mv.data)
        x = np.array(self.x_mv.data).reshape((-1, 1))
        self.model.fit(x, y)

    def is_ready(self):
        return len(self.x_mv.data) >= 3

    def predict(self, target: float):
        return self.model.predict(np.asarray([target]).reshape((-1, 1)))[0]



import unittest

from xh_cpu_usage_simulator.MovingAvg import MovingAvg


class TestMovingAvg(unittest.TestCase):
    def test_mv(self):
        m_avg = MovingAvg(3)
        m_avg.insert(6)
        self.assertEquals(2, m_avg.avg)
        m_avg.insert(1)
        self.assertEquals(7/3, m_avg.avg)
        self.assertEquals(6, m_avg.data[0])
        self.assertEquals(1, m_avg.data[1])
        self.assertEquals(0, m_avg.data[2])

        m_avg.insert(3)
        self.assertEquals(10/3, m_avg.avg)
        self.assertEquals(6, m_avg.data[0])
        self.assertEquals(1, m_avg.data[1])
        self.assertEquals(3, m_avg.data[2])

        m_avg.insert(4)
        self.assertEquals(8/3, m_avg.avg)
        self.assertEquals(4, m_avg.data[0])
        self.assertEquals(1, m_avg.data[1])
        self.assertEquals(3, m_avg.data[2])



from dataclasses import dataclass
from typing import Generator

import psutil


@dataclass
class Metrics:
    avg: float


class PsMonitor:

    def monitor(self) -> Generator[Metrics, None, None]:
        cpu_count = psutil.cpu_count()
        while True:
            result = psutil.cpu_percent(interval=1, percpu=True)
            yield Metrics(sum([i / cpu_count for i in result]))



import time

from xh_cpu_usage_simulator.MovingAvg import MovingAvg


def worker(index: int, d: dict, r: dict, do_something) -> [int, int]:
    def get_time():
        t = time.time_ns()
        ms = t // 1_000_000
        s = ms // 1000
        return s, ms

    s, _ = get_time()

    task_to_run = int(d[f"{index}"])
    target = task_to_run
    mv = MovingAvg(10)
    while True:
        n_s, n_ms = get_time()
        if n_s != s:
            new_task_to_run = int(d[f"{index}"])
            # if task_to_run != new_task_to_run:
            #     print(f"[Worker {index}]Target to run updated from {task_to_run} -> {new_task_to_run}")
            # if target != 0:
            #     print(f"[Worker {index}] Completed {task_to_run - target}/{task_to_run} [{round(1- target/task_to_run, 2)} %]")

            completed = task_to_run - target
            mv.insert(completed)
            r.update({f"{index}-complete": round(mv.avg,2)})

            task_to_run = new_task_to_run
            # print(f"[Worker {index}]Rest target")
            target = task_to_run
            s = n_s

        if target > 0:
            target -= 1
            do_something()  # payload
        elif target == 0:
            diff = (s + 1) * 1000 - n_ms
            # print(f"[Worker {index}]Taking {(n_ms - (s * 1000))} ms for {task_to_run} tasks [Sleep for {round(diff/1000, 4)} s]")
            if diff > 0:
                time.sleep(diff/1000)
            else:
                continue



from typing import Optional


class MovingAvg():
    def __init__(self, size: int):
        self.data = []
        self.size = size
        self.cur_index = 0
        self.avg = 0
        # for _ in range(size):
        #     self.data.append(0)

    def insert(self, value: float):
        if len(self.data) < self.size:
            self.data.append(value)
        else:
            self.data[self.cur_index] = value
        len_of_data = len(self.data)
        total_data = sum(self.data)
        self.avg = round(total_data / len_of_data, 2)
        self.cur_index = (self.cur_index + 1) % len(self.data)

    def __repr__(self):
        return f"{len(self.data)} - {self.avg}"

    def __str__(self):
        return self.__repr__()



# This is a sample Python script.
import argparse
import sys
from threading import Thread

import psutil

from xh_cpu_usage_simulator import TaskWorker
from xh_cpu_usage_simulator.MovingAvg import MovingAvg
from xh_cpu_usage_simulator.PsMonitor import PsMonitor
from xh_cpu_usage_simulator.LinearRegressionPrediction import LinearRegressionPredication


def validation(min_usage: float, max_usage: float):
    if not 0 < min_usage < 100:
        raise Exception(f"minimum usage[{min_usage}] is out of range")
    if not 0 < max_usage < 100:
        raise Exception(f"maximum usage[{max_usage}] is out of range")

    if max_usage < min_usage:
        raise Exception(f"Usage bound [{min_usage} - {max_usage}] is not valid")


if __name__ == '__main__':
    parser = argparse.ArgumentParser(
        prog="xh-cpu-usage-simulator",
        usage="A simple tool to simulate system loading of cpu"
    )

    parser.add_argument("--min", help="minimum percentage cpu usage", type=float)
    parser.add_argument("--max", help="maximum percentage cpu usage", type=float)

    parsed_args = parser.parse_args(sys.argv[1:])

    lower_bound = parsed_args.min
    upper_bound = parsed_args.max

    validation(lower_bound, upper_bound)

    d = dict()
    r = dict()
    for i in range(psutil.cpu_count(logical=False)):
        d.update({f"{i}": 10})
        r.update({f"{i}-complete": 0})
        bucket = MovingAvg(10)
        Thread(target=TaskWorker.worker, args=[i, d, r, lambda: bucket.insert(1)]).start()

    total_operation = sum([d[i] for i in d])

    m_avg = MovingAvg(10)

    # target = 70
    rounds = 0
    prediction = LinearRegressionPredication(10)
    for m in PsMonitor().monitor():
        if m_avg.cur_index == 0:
            if rounds == 0:
                rounds += 1
            else:
                rounds += 1
                cur_total_operation = round(sum([d[i] for i in d]), 2)
                cur_total_mv = round(sum([r[i] for i in r], 2))
                prediction.insert(m_avg.avg, cur_total_mv)

                target = round(lower_bound+abs(lower_bound-upper_bound)/2,2)

                def next_operation():
                    if prediction.is_ready():
                        return prediction.predict(target)
                    else:
                        return None


                if lower_bound > m_avg.avg:
                    for i in d:
                        n_op = next_operation()
                        d[i] = n_op/len(d) if n_op is not None and n_op > 0 else d[i] * 2
                    new_total_operation = round(sum([d[i] for i in d]), 2)
                    print(
                        f"{rounds:05d}::Targeting[{lower_bound} < {target} < {upper_bound}], now[{m_avg.avg}], increase operation with loading[{cur_total_mv}] from {cur_total_operation} to {new_total_operation}")
                elif upper_bound < m_avg.avg:
                    for i in d:
                        n_op = next_operation()
                        d[i] = n_op/len(d) if n_op is not None else d[i] * 0.85
                    new_total_operation = sum([d[i] for i in d])
                    print(
                        f"{rounds:05d}Targeting[{lower_bound} < {target} < {upper_bound}], now[{m_avg.avg}], decrease operation with loading[{cur_total_mv}] from {cur_total_operation} to {new_total_operation}")
                else:
                    for i in d:
                        n_op = next_operation()
                        d[i] = n_op/len(d) if n_op is not None else d[i] * 0.85
                    new_total_operation = sum([d[i] for i in d])
                    print(
                        f"{rounds:05d}Targeting[{lower_bound} < {target} < {upper_bound}], now[{m_avg.avg}], adjustment with loading[{cur_total_mv}] from {cur_total_operation} to {new_total_operation}")
        m_avg.insert(m.avg)

# See PyCharm help at https://www.jetbrains.com/help/pycharm/
